/**
  * ClockLet.java
  * 
  * A complete MPEGlet using the Scene API to dynamically control an
  * otherwise static MPEG-4 clock scene. This software module was originally 
  * developed by IBM Research (T.J.Watson) in the course of development 
  * of the MPEG-4 Systems (ISO/IEC 14496-1) standard and is featured in
  * this publication courtesy of IBM and the authors of this code.
  * @author Steve Wood, IBM  TJ Watson Research, USA
  * @author Dr. Michelle Kim, IBM  TJ Watson Research, USA 
  */

import org.iso.mpeg.mpegj.*;
import org.iso.mpeg.mpegj.scene.*;
import java.util.Calendar;

/** This MPEGlet to support IBMMpegjClock
 */
public class ClockLet implements MPEGlet, SceneListener, SFFloatFieldValue,
                                 MFStringFieldValue, SFStringFieldValue
{
  private SceneManager mgr;
  private Scene        scene;
  private int          state;  // 0 = running, 1 = stopping, 2 = stopped

  private Node    nodeSecondHand;
  private float   seconds;

  private Node    nodeMinuteHand;
  private float   minutes;

  private Node    nodeHourHand;
  private float   hours;

  private Node     nodeDayText;
  private int      day    = -1;
  private String[] days   = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};

  private Node     nodeDateText;
  private int      date    = -1;

  private Node     nodeMonthText;
  private int      month  = -1;
  private String[] months = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", 
                             "Aug", "Sep", "Oct", "Nov", "Dec"};

  // This table represents node values after examining the resulting BIFS.
  static private final int dayTextDEF    = 4;
  static private final int dateTextDEF   = 6;
  static private final int monthTextDEF  = 7;
  static private final int minuteHandDEF = 8;
  static private final int hourHandDEF   = 9;
  static private final int secondHandDEF = 10;

  /** (MPEGlet)
   * Cleanup operations to perform before the MPEGlet is destroyed go here.
   */
  public void destroy()
  {
  }

  /** (MPEGlet)
   * Operations to perform before the MPEGlet runs go here.
   */
  public void init()
  {
  }

  /** (SceneListener)
   * Called by the scene manager when the BIFS scene has
   * been changed.
   *
   * @param i contains the message number that indicates the change.
   * @param scene contains the new Scene object when appropriate.
   */
  public synchronized void notify(int i, Scene scene)
  {
    if (   i == SceneListener.Message.SCENE_READY
        || i == SceneListener.Message.SCENE_REPLACED
       )
    {
      this.scene = scene;
      notify();
    }
  }

  /** (MPEGlet/Runnable) 
   * Operations necessary for the MPEGlet's main thread of execution go here. 
   */
  public void run()
  {
    try
    {
      mgr = (new MpegjTerminal()).getSceneManager();
      mgr.addSceneListener(this);
      synchronized(this)
      {
          // Wait until SCENE_READY
          while(scene == null)
              wait();
      }

      Thread.sleep(1000);  // We don't need a delay but it looks nicer

      nodeSecondHand = scene.getNode(secondHandDEF);
      nodeMinuteHand = scene.getNode(minuteHandDEF);
      nodeHourHand   = scene.getNode(hourHandDEF);

      nodeDayText    = scene.getNode(dayTextDEF);
      nodeDateText   = scene.getNode(dateTextDEF);
      nodeMonthText  = scene.getNode(monthTextDEF);

      stringsValue[0] = this;

      while (state == 0)
      {
        int secCount = -1;
        Thread.sleep(50);

        Calendar now = Calendar.getInstance();

        if (secCount != now.get(Calendar.SECOND))
        {
          secCount   = now.get(Calendar.SECOND);

          float secs = now.get(Calendar.SECOND);
          float mins = now.get(Calendar.MINUTE);
          float hrs  = now.get(Calendar.HOUR);
          
          seconds = -(float)(((secs            )/60.0) * Math.PI * 2.0);
          minutes = -(float)(((mins + secs/60.0)/60.0) * Math.PI * 2.0);
          hours   = -(float)(((hrs  + mins/60.0)/12.0) * Math.PI * 2.0);
          
          floatValue = seconds;
          nodeSecondHand.sendEventIn(EventIn.Transform2D.rotationAngle, this);
          
          floatValue = minutes;
          nodeMinuteHand.sendEventIn(EventIn.Transform2D.rotationAngle, this);
          
          floatValue = hours;
          nodeHourHand.sendEventIn(EventIn.Transform2D.rotationAngle,   this);
        }

        if (day != now.get(Calendar.DAY_OF_WEEK))
        {
          day = now.get(Calendar.DAY_OF_WEEK);
          stringValue = days[day-1];
          nodeDayText.sendEventIn(EventIn.Text.string, this);
        }

        if (date != now.get(Calendar.DATE))
        {
          date = now.get(Calendar.DATE);
          stringValue = Integer.toString(date);
          nodeDateText.sendEventIn(EventIn.Text.string, this);
        }

        if (month != now.get(Calendar.MONTH))
        {
          month = now.get(Calendar.MONTH);
          stringValue = months[month];
          nodeMonthText.sendEventIn(EventIn.Text.string, this);
        }
      }

      stopped();
    }
    catch(MPEGJException mpegjexception)
    {
        stopped();
        System.out.println("ClockLet MPEGJ exception");
        throw new Error(String.valueOf(mpegjexception.getClass()) + ": " + 
                        mpegjexception.getMessage());
    }
    catch(InterruptedException interruptedexception)
    {
        stopped();
        System.out.println("ClockLet interrupted exception");
        throw new Error(String.valueOf(interruptedexception.getClass()) + ": " + 
                        interruptedexception.getMessage());
    }
    catch (Exception rawexception)
    {
        stopped();
        System.out.println("ClockLet RAW exception");
        throw new Error(String.valueOf(rawexception.getClass()) + ": " + 
                        rawexception.getMessage());
    }
  }

  private synchronized void stopped()
  {
    state = 2;
    notify();
  }

  /** (MPEGlet)
   * Operations to perform before the MPEGlet is stopped go here.
   */
  public synchronized void stop()
  {
      try {
          mgr.removeSceneListener(this);
      } catch (MPEGJException mpegjexception) {
          System.out.println(mpegjexception);
      }

      if (state != 2) {
          state = 1;  // 1 = stopping

          // Wait until stopped
          while(state != 2) {
              try {
                  wait();
              } catch (InterruptedException ex) {
                  return;
              }
          }
      }
  }

  private float floatValue;
  public float getSFFloatValue() {return floatValue;}

  private SFStringFieldValue[] stringsValue = new SFStringFieldValue[1];
  public SFStringFieldValue[] getMFStringValue() {return stringsValue;}

  private String stringValue;
  public String getSFStringValue() {return stringValue;}
}
