#  $Id: LayeredFrame.py,v 2.9.1.3 2001-08-14 14:40:09-05 brinkman Exp $
#  A subclass of PowerFrame that manages various layers of contents.
#  Copyright (C) 2000  Peter Brinkmann (brinkman@math.uiuc.edu)
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
LayeredFrame.py: A frame that manages one or more layers.

$Id: LayeredFrame.py,v 2.9.1.3 2001-08-14 14:40:09-05 brinkman Exp $
"""

import Config
from PowerFrame import PowerFrame

###############################################################################

class LayeredFrame(PowerFrame):
	"""
$Id: LayeredFrame.py,v 2.9.1.3 2001-08-14 14:40:09-05 brinkman Exp $

The class LayeredFrame manages a list of dependent objects (layers), each
of which must contain some data as well as a menu bar. Only one layer (and
its menu bar) is displayed at a time, and there are navigation buttons that
allow users to move back and forth between the layers. The purpose of this
class is to preserve real estate on the screen.

More precisely, the layers must be classes with the following properties:

	1. They must be subclasses of MenuMaker.

	2. They must have a constructor of the form
		 __init__(self,master,data),
	   where master will be the ambient frame, and the data is given
	   for initialization purposes (possibly None).

	3. They must have a string variable self.name, and they must NOT
	   have an entry self.label because this one will be assigned by
	   LayeredFrame for reference purposes.

	4. All their components (except for the menu bar) must be contained
	   in one panel called self.panel.

See the module TextLayer.py for a good example.
"""

	version="""
$Id: LayeredFrame.py,v 2.9.1.3 2001-08-14 14:40:09-05 brinkman Exp $
(c) 2000 Peter Brinkmann (brinkman@math.uiuc.edu)
"""
	helptext="""
Click on the navigation arrows or 'Go' button to
display another layer.
"""

###############################################################################

	def __init__(self,cls,dat='',name='LayeredFrame'):
		"""
Constructor for LayeredFrame.

Parameters:

	- cls: a class (NOT an instance!) that is suitable as a layer

	- dat: initial data for this class
"""
		self.index=0
		self.entries=[]
		self.count=0
		PowerFrame.__init__(self,None,name)
		self.addInternalDependent(cls,dat)
		self.setVisible(1)
		self.pack()

###############################################################################

	def focusOn(self,dep):
		"""
Brings a given layer to the front.

Parameters:

	- dep: dependent object to be displayed on top.
"""
		self.setAllCursorsBusy()
		self.index=self.entries.index(dep)
		self.setMenuBar(dep.menubar)
		self.cl.show(self.panel,dep.label)
		self.setTitle(dep.name)
		self.resetAllCursors()
		self.updateWidgets()

###############################################################################

	def addInternalDependent(self,cls,dat=None):
		"""
Adds a new layer.

Parameters:

	- dep: class (NOT instance) of layer to be added

	- dat: initial data for dependent

Returns:

	- new dependent (usually, you can simply ignore the return value)

Note that there's *no* method called removeInternalDependent. In order to
remove an internal dependent, simply call removeDependent; LayeredFrame
is smart enough to figure out the rest.
"""
		dep=cls(self,dat)
		try:
			dep.label=('%d: ' % self.count)+dep.name
		except:
			dep.label=('%d: ' % self.count)+_('No name.')
		self.count=self.count+1
		if len(self.entries)==0:
			self.entries.append(dep)
		else:
			self.entries.insert(self.index+1,dep)
		self.panel.add(dep.panel,dep.label)
		self.focusOn(dep)
		return dep

###############################################################################

	def removeDependentHook(self,dep):
		"""
Overrides the corresponding method of MenuMaker to handle the additional
book-keeping for internal dependents.

Parameters:

	- dep: dependent to be removed

If dep is the last layer to be removed, then the frame closes itself.
"""
		self.updateWidgets()
		if dep in self.entries:
			if len(self.entries)<2:
				self.entries.remove(dep)
				self.close()
				return
			else:
				ind=self.entries.index(dep)
				if self.index!=ind:
					nf=self.entries[self.index]
				elif self.index<len(self.entries)-1:
					nf=self.entries[self.index+1]
				else:
					nf=self.entries[self.index-1]
				self.focusOn(nf)
				self.entries.remove(dep)
				self.index=self.entries.index(nf)
				self.updateWidgets()
				self.cl.removeLayoutComponent(self.panel,
					dep.label)
				self.remove(dep.panel)

###############################################################################

	def previous(self,e=None):
		"""
Switches to the layer before the current one.
"""
		if self.index>0:
			self.focusOn(self.entries[self.index-1])

###############################################################################

	def next(self,e=None):
		"""
Switches to the layer after the current one.
"""
		if self.index<len(self.entries)-1:
			self.focusOn(self.entries[self.index+1])

###############################################################################

	def goTo(self,e=None):
		"""
Displays a dialog that allows the user to chose the new focus.
"""
		ll=[]
		dict={}
		CANCEL=_('Cancel')
		for i in self.entries:
			ll.append((i.label,lambda:None))
			dict[i.label]=i
		ll.append((CANCEL,lambda:None,1))
		ans=self.question(self.name,_('Choose entry.'),ll,1)
		if ans!=CANCEL and ans!=None:
			self.focusOn(dict[ans])

###############################################################################
# The remaining methods are for internal use only.                            #
###############################################################################

	def start(self):
		self.PREV='<-- '+_('Prev')
		self.NEXT=_('Next')+' -->'
		self.GO=_('Go')
		self.tBar=[('S',[(self.PREV,self.previous),
						(self.GO,self.goTo),
						(self.NEXT,self.next)])]


	def makeWidgets(self):
		self.cl=Config.factory.constructCardLayout()
		self.panel=Config.factory.constructPanel(self.cl)
		self.add(self.panel,Config.factory.locations['C'])


	def updateWidgets(self):
		self.buttons[self.PREV].setEnabled(self.index>0)
		self.buttons[self.GO].setEnabled(len(self.entries)>1)
		self.buttons[self.NEXT].setEnabled(
				self.index<len(self.entries)-1)

###############################################################################

class OneLayer(LayeredFrame):
	"""
Just a simple class that does everything that LayeredFrame does, except it
only allow one layer. The idea is to write code for layers and be able to
use them as single layers, without the navigation arrows of LayeredFrame.
"""

	def start(self):
		pass

	def updateWidgets(self):
		pass

###############################################################################


