#  $Id: MenuMaker.py,v 2.26.1.3 2001-08-14 18:14:58-05 brinkman Exp $
#  A class that provides high level GUI features.
#  Copyright (C) 2000  Peter Brinkmann (brinkman@math.uiuc.edu)
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
MenuMaker.py: The heart and soul of the package.

$Id: MenuMaker.py,v 2.26.1.3 2001-08-14 18:14:58-05 brinkman Exp $
"""


import Config
import string
import sys

###############################################################################

class MenuMaker:
	"""
$Id: MenuMaker.py,v 2.26.1.3 2001-08-14 18:14:58-05 brinkman Exp $

This class generates certain high-level GUI features, such as complete
menu bars, tool bars, predefined dialogs for various purposes, etc.
MenuMaker is not intended for stand-alone use; it needs to be subclassed
and mixed with frames, for example (for example, the class PowerFrame mixes
MenuMaker and Frame).

An instance of MenuMaker has a list of dependent objects, as well as a
pointer to its owner. MenuMaker also has a static flag (MenuMaker.isApplet)
that indicates whether instances belong to an applet (default is 0, applets
need to say MenuMaker.isApplet=1 in their constructor).

Default fonts and colors are maintained by the widget factory.
These variables can be overridden by subclassing the factory.

Those functions that are meant to be called from other modules are documented,
the rest is for internal use only.

Inspired by (but much different from and much more powerful than) the class
GuiMixin from Mark Lutz, Programming Python, O'Reilly.
"""


	version="""
$Id: MenuMaker.py,v 2.26.1.3 2001-08-14 18:14:58-05 brinkman Exp $
(c) 2000 Peter Brinkmann (brinkman@math.uiuc.edu)
"""
	helptext="""
Fill in the main help text of your class here.
"""

	separator='__separator__'
	checkBoxMarker='_'
	isApplet=0
	echoChar='.'	# echo character for entering passwords

###############################################################################

	def __init__(self,ms,isframe,name='MenuMaker'):
		"""
Constructor for MenuMaker

Parameters:

	- ms: the master of this instance, usually either None or the object
	    that owns this instance (frequently, this object will also be
	    of type MenuMaker)

	- isframe: a flag that indicates whether this instance is a subclass of
		 Frame.
"""
		# the following seven definitions are slightly redundant,
		# but convenient.
		self.defaultfont=Config.factory.defaultfont
		self.dialogfont=Config.factory.dialogfont
		self.filedialogfont=Config.factory.filedialogfont
		self.textfont=Config.factory.textfont
		self.textfontlarge=Config.factory.textfontlarge

		self.dialogbackcolor=Config.factory.dialogbackcolor
		self.dialogtextcolor=Config.factory.dialogtextcolor

		self.mBar=[]
		self.tBar=[]
		self.master=ms
		self.name=name
		self.dependentsList=[]
		self.flashBoxStack=[]
		self.cursorStack=[]
		try:
			self.master.addDependent(self)
		except:
			pass
		self.isFrame=isframe
		if self.isFrame:
			self.frame=self
		else:
			self.frame=self.master.frame
		self.start()
		self._makeDefaultMenuBar()
		self._makeDefaultToolBars()
		self.makeWidgets()

###############################################################################

	def start(self):
		"""
Method called by constructor.

This method needs to contain the definition of
the variables self.mBar and self.tBar, for example:

                self.mBar=[
                        ('File',[('New',[('Worksheet',self.newWorksheet,{}),
																		^
												# A dictionary at the end
												# of a tuple for a MenuItem
												# can contain additional
												# keyword arguments for the
												# constructor of MenuItem.
												# This might look like
												# {'icon':someIcon}.
                                         ('Copy',self.newCopy)
                                        ]
                                 ),		# cascading menus!!
                                 self.separator,
                                 ('Save',self.save,1),
						   ^
					# the (optional) entry 1 indicates
					# that this menu item is enabled
					# 'enabled' is the default
					# 0 would indicate that it is
					# disabled
                                 self.separator,
                                 ('Close',self.close,1,'c'),
							^
					# this (optional) entry stands for the
					# keyboard accelerator Ctrl-c
					# the enabled-flag is required when
					# using keyboard accelerators
                                 ('Exit',self.exit,1,'X')
						      ^
					# keyboard accelerator Ctrl-Shift-x
                                ],
                                1,      # enabled
                                1       # tear-off
					# the last two flags are optional.
					# (the default is enabled/not tear-off)
					# However, if the menu is supposed to
					# tear-off, then the enabled flag is
					# required.
                        ),
                        ('View',[('Clear',self.clear),
                                 self.separator,
				 # here's how to specify checkboxes!
                                 (self.checkBoxMarker+'Disk model',
                                                self.repaint2,1,1)
				 #			      ^^^
				 # those two flags mean enabled, initially set
				 # to 'checked'
                                ]
                        ),
                        ('Help',[('Help window',self.help),
                                 ('About',self.about)
                                ]
                        )
                ]
                self.tBar=[('S',
		#	    ^ location of tool bar
		#		('N','S','E','W')
		#
				[('Prev',self.previous,1),
				 ('Go',self.goTo,0),
				 ('Next',self.next)]
		#	    	list of buttons with
		#		label and function
		# The description of a button is of the form
		# (label,function[,enabled])
		#
			   ),
			   ('N',[('Another_button',self.someMethod),
				 ('Yet_another_button',self.anotherMethod,{})]
														  ^
					# an (optional) dictionary at the end of a tuple
					# can contain additional keyword parameters for the
					# constructor of the button, e.g., an icon (for JButton)
					# This might look like {'icon':someIcon}
			   )
		]


Note: You can use either lists or tuples in the definition of self.mBar and
self.tBar; it doesn't make any difference.

The functions in the lists need to have one dummy parameter (for the Java
event) that's usually ignored by the method. My functions for this purpose
usually look like this example:

	def reactToSomeMenu(self,e=None):
		# do some stuff...


The instances will have three dictionaries for buttons and menu items:

	- self.menuitems: keys are labels of items

	- self.toolbars: keys are the locations of the tool bars, i.e., 'N',
		'S', 'E', and 'W' (for displaying the toolbars, the location
		has to be translated by the dictionary Config.factory.locations,
		e.g., Config.factory.locations['N'] equals
		BorderLayout.NORTH (for the default factory))

	- self.buttons:	keys are the labels of buttons

These dictionaries are useful if you want to disable/enable widgets
dynamically.
"""
		pass

###############################################################################

	def makeMenuBar(self,mBar):
		"""
Method for generating additional menu bars.

Parameters:

	- mBar:	description of the menu bar, same format as self.mBar
		in the method start()

Returns:

	- tuple: the resulting menubar, a dictionary containing the various
		menus and menu items
"""
		menuitems={}
		menubar=Config.factory.constructMenuBar()
		for i in mBar:
			menubar.add(self._makeMenu(i,menuitems,0))
		return menubar,menuitems

###############################################################################

	def makePopupMenu(self,mlist):
		"""
Method for generating pop up menus.

Parameters:

	- mlist: description of pop up menu in the following format. Example:

                        ('Help',[('Help window',self.help),
                                 ('About',self.about)
                                ]
                        )

Returns:

	- tuple: the pop up menu, a dictionary containing the menu items
"""
		menuitems={}
		pumenu=self._makeMenu(mlist,menuitems,1)
		return pumenu,menuitems

###############################################################################

	def makeToolBars(self,tBar):
		"""
Method for generating additional tool bars.

Parameters:

	- tBar:	description of the tool bars, same format as self.tBar
		in the method start()

Returns:

	- tuple: a dictionary containing the resulting tool bars (a tool bar
		is a Panel with a number of buttons), a dictionary containing
		the various buttons

The tool bars are instances of Panel, so you can
immediately add them to your layouts
"""
		buttons={}
		toolbars={}
		if len(tBar)>0:
			for j in tBar:
				if j[0] in ['N','S']:
					toolbar=Config.factory.constructPanel(
					   Config.factory.constructFlowLayout())
				else:
					toolbar=Config.factory.constructPanel(
					   Config.factory.constructGridLayout(
						len(j[1]),1))
				for i in j[1]:
					if type(i[-1])!=type({}):
						kwargs={}
					else:
						kwargs=i[-1]
						i=i[:-1]
					kwargs['actionPerformed']=i[1]
					b=apply(Config.factory.constructButton,(i[0],),kwargs)
					if len(i)>2:
						b.setEnabled(i[2])
					buttons[i[0]]=b
					toolbar.add(b)
				toolbars[j[0]]=toolbar
		return toolbars,buttons

###############################################################################

	def addDependent(self,dep):
		"""
Adds a new dependent to the list. Calls self.addDependentHook, which needs
to be overridden when the addition of a new dependent requires extra care.
"""
		if not dep in self.dependentsList:
			self.dependentsList.append(dep)
			self.addDependentHook(dep)

###############################################################################

	def addDependentHook(self,dep):
		"""
Hook for extra bookkeeping when adding a dependent. To be overridden by
subclasses.
"""
		pass

###############################################################################

	def removeDependent(self,dep):
		"""
Removes a dependent from the list. Calls self.removeDependentHook, which needs
to be overridden when the removal of a dependent requires extra care.
"""
		if dep in self.dependentsList:
			self.dependentsList.remove(dep)
			self.removeDependentHook(dep)

###############################################################################

	def removeDependentHook(self,dep):
		"""
Hook for extra bookkeeping when removing a dependent. To be overridden
by subclasses.
"""
		pass

###############################################################################

	def makeWidgets(self):
		"""
If your subclass has widgets other than menus and tool bars, then you can
override this class to create the widgets here. The constructor of MenuMaker
calls this method.
"""
		pass

###############################################################################

	def updateWidgets(self):
		"""
You need to override this method if you have widgets that change dynamically.
"""
		pass

###############################################################################

	def finish(self):
		"""
Hook for additional cleanup work before closing. Should return 1 iff
cleanup was successful and closing can continue.
"""
		return 1

###############################################################################

	def focusOn(self,dep):
		"""
Hook for focussing on a dependent.
"""
		pass

###############################################################################

	def close(self,e=None):
		"""
Method for closing (closes this instance as well as its dependents, aborts
closing when one of the finish methods returns 0). Usually doesn't need
overriding because additional functionality can be added in self.finish().
"""
		ll=self.dependentsList[:]
		for dep in ll:
			if dep in self.dependentsList:
				self.focusOn(dep)
				if not dep.close():
					return 0
		if not self.finish():
			return 0
		try:
			self.master.removeDependent(self)
		except:
			pass
		return 1

###############################################################################

	def closeFrame(self,e=None):
		"""
Closes the ambient frame.
"""
		self.frame.close()

###############################################################################

	def exit(self,e=None):
		"""
Exits altogether.
"""
		self.frame.exit()

###############################################################################

	def question(self,title,text,opts,modal=1):
		"""
A dialog with a number of buttons.
Parameters:

	- title: the title string

	- text: the text of the main label

	- opts: list of buttons and functions to be called, for example:

		[('button 1',self.function1),('button 2',self.function2,1)]
									^
							this flag indicates
							that 'button 2' is
							supposed to have the
							focus when this dialog
							is shown
							(not implemented)

	- modal: flag indicating whether the dialog is supposed to be modal

Return value:

	- the label of the selected button in the modal case

	- a pointer to this dialog in the non-modal case

Also, the label of the selected button can be found in the variable
this.reply. The master object can set the font of the dialog by changing the
value of the variable Config.factory.dialogfont.
"""
		d=Config.factory.constructQDialog(
			self,self.frame,title,text,opts,modal)
		d.show()
		if modal:
			return d.reply
		else:
			return d

###############################################################################

	def inputString(self,title,text,value=None,
					columns=None,submit=None,cancel=None):
		"""
A dialog (always modal) with a TextField for entering Strings.

Parameters:

	- title: the title of the dialog

	- text: the text indicating what the user is supposed to enter

	- value: the initial value of the TextField (can be None)

	- columns: the size of the TextField (can be None, in which case
		the size will be the length of value plus five if value!=None,
		20 otherwise)

	- submit: Label of the submit button (can be None, in which case the
		button won't be there)

	- cancel: Label of the cancel button (can be None, in which case the
		button won't be there)

Return value:

	- the value of the TextField if the user hits 'Enter' or clicks on
	  the submit button

	- None if the user clicks on the cancel button
"""
		d=Config.factory.constructTDialog(
			self,self.frame,title,text,value,columns,
			submit,cancel,None)
		d.show()
		return d.value

###############################################################################

	def inputPassword(self,title,text,
			value=None,columns=None,submit=None,cancel=None):
		"""
Like inputString, except the input string will be replaced by the character
self.echoChar (by default, self.echoChar is '.', but this can be overridden).
"""
		d=Config.factory.constructTDialog(
			self,self.frame,title,text,value,columns,
			submit,cancel,self.echoChar)
		d.show()
		return d.value

###############################################################################

	def infoBox(self,title,text,btext=None,modal=1):
		"""
A simple dialog box that shows a message and asks the user to acknowledge.

Parameters:

	- title: the title string

	- text: the text of the main label

	- btext: label of the button, defaults to 'OK'

	- modal: default is 1

Moreover, the master object can set the font of the dialog by changing the
value of the variable Config.factory.dialogfont.
"""
		if btext==None:
			btext=_('OK')
		d=Config.factory.constructQDialog(self,self.frame,title,text,
				[(btext,lambda:None)],modal)
		d.show()

###############################################################################

	def flashBox(self,title,text):
		"""
A simple message box that expects no action from the user
(like 'computation in progress...'). Instances of MenuMaker maintain a
stack of flash boxes belonging to them.

Parameters:

	- title: the title string

	- text: the message string

Moreover, the master object can set the font of the dialog by changing the
value of the variable Config.factory.dialogfont.

Returns:

	- a pointer to this instance if the ambient frame is visible

	- None otherwise
"""
		if self.frame.isVisible():
			d=Config.factory.constructQDialog(
				self,self.frame,title,text,[],0)
			d.show()
			self.flashBoxStack.append(d)
			return d
		else:
			return None

###############################################################################

	def updateFlashBox(self,d,title,text):
		"""
Updates the title and message of a given flash box.

Parameters:

	- d: given flash box

	- title: new title

	- text: new message
"""
		if d!=None:
			d.setTitle(title)
			d.setLabel(text)
			d.show()
			d.repaint()

###############################################################################

	def closeLastFlashBox(self):
		"""
Closes the most recent flash box (recommended way of getting rid of flash
boxes).
"""
		try:
			d=self.flashBoxStack[-1]
			self.flashBoxStack.remove(d)
			d.close()
		except:
			pass

###############################################################################

	def closeFlashBox(self,d):
		"""
Closes the given flash box. This messes up the stack of flash boxes, so 
make sure you keep track of your flash boxes when you use this.

Parameters:

	- d: given flash box
"""
		if d in self.flashBoxStack:
			self.flashBoxStack.remove(d)
		d.close()

###############################################################################

	def errorBox(self,text):
		"""
Displays an error message and waits for the user to acknowledge.

Parameters:

	- text: error message

Moreover, the master object can set the font of the dialog by changing the
value of the variable Config.factory.dialogfont.
"""
		self.infoBox(_('Exception'),text,_('OK'),1)

###############################################################################

	def exceptionBox(self):
		"""
Displays an error message extracted from the most recent exception and
waits for the user to acknowledge (expects class based exceptions derived
from the standard Exception class).
"""
		self.errorBox(str(sys.exc_info()[0])+'\n'+
				str(sys.exc_info()[1]))

###############################################################################

	def loadDialog(self,title,func):
		"""
A simple load dialog.

Parameters:

	- title: title of FileDialog

	- func: method with one string parameter (the file name) that does
		the loading

Moreover, the master object can set the font of the dialog by changing the
value of the variable Config.factory.filedialogfont.
"""
		return Config.factory.loadDialog(self,title,func)

###############################################################################

	def saveDialog(self,title,func):
		"""
A simple save dialog.

Parameters:

	- title: title of FileDialog

	- func: method with one string parameter (the file name) that does
		the saving

Moreover, the master object can set the font of the dialog by changing the
value of the variable Config.factory.filedialogfont.
"""
		return Config.factory.saveDialog(self,title,func)

###############################################################################

	def help(self,e=None):
		"""
Displays a window with the standard help string (self.helptext).
Assign a new value to self.helptext to use in your own subclass.
"""
		self.specialHelp(self.name+' '+_('Help'),
				self.helptext)

###############################################################################

	def specialHelp(self,title,text):
		"""
Displays text in a window, usually as a help feature.

Parameters:

	- title: title string

	- text: help text

"""
		try:
			self.frame.showText(self,title,text)
		except:
			self.errorBox(_('Help feature not available.'))

###############################################################################

	def displayPic(self,title,img):
		"""
Displays a window with an image.

Parameters:

	- title: title string

	- img: string containing an image (e.g., in .jpeg format)
"""
		self.frame.showPic(self,title,img)

###############################################################################

	def about(self,e=None):
		"""
Shows a dialog with the version string (self.version), plus possibly a
copyright message, etc.
"""
		self.infoBox(_('Version'),self.version,_('OK'),1)

###############################################################################

	def getAllCursors(self):
		"""
Returns a dictionary where the keys are the components of this instance, and
the values are the current mouse cursors of those components.
"""
		oc={}
		ll=self._getAllComponents()
		for i in ll:
			oc[i]=i.getCursor()
		return oc

###############################################################################

	def setAllCursors(self,cs):
		"""
Sets the mouse cursors of all components of this instance to one given value.
An instance of MenuMaker has a list of previous cursors, so you can undo this
by calling self.resetAllCursors().

Parameters:

	- cs: new cursor

Returns:

	- list of all cursors before the change.
"""
		oc=self.getAllCursors()
		for i in oc.keys():
			i.setCursor(cs)
		self.cursorStack.append(oc)
		return oc

###############################################################################

	def resetAllCursors(self):
		"""
Sets the cursors of all components of this instance back to the state at the
top of the stack of cursor values.
"""
		try:
			oc=self.cursorStack[-1]
			self.cursorStack.remove(oc)
			self.setCursorsFromDict(oc)
		except:
			pass

###############################################################################

	def setCursorsFromDict(self,oc):
		"""
Sets all cursors according to values from a given dictionary.

Parameters:

	- oc: dictionary in the form returned by self.getAllCursors()
"""
		for i in oc.keys():
			i.setCursor(oc[i])

###############################################################################

	def setAllCursorsBusy(self):
		"""
Replaces the cursors of all components of this instance by the wait cursor.
Can be undone with self.resetAllCursors().
"""
		return self.setAllCursors(
			Config.factory.constructWaitCursor())


###############################################################################
# from this point on, most of the methods are intended for internal use only. #
###############################################################################

	def _makeDefaultMenuBar(self):
		self.menubar,self.menuitems=self.makeMenuBar(self.mBar)


	def _makeMenu(self,i,menuitems,popup=0):
		if type(i[-1])!=type({}):
			kwargs={}
		else:
			kwargs=i[-1]
			i=i[:-1]
		if popup:
			menu=apply(Config.factory.constructPopupMenu,(i[0],),kwargs)
		elif len(i)>3:
			menu=apply(Config.factory.constructMenu,(i[0],i[3]),kwargs)
		else:
			menu=apply(Config.factory.constructMenu,(i[0],),kwargs)
		for j in i[1]:
			if j==self.separator:
				menu.addSeparator()
			else:
				if type(j[1])==type([]) or\
				   type(j[1])==type((0,)):
					item=self._makeMenu(j,menuitems,0)
				else:
					if type(j[-1])!=type({}):	# handle additional options
						kwargs={}
					else:
						kwargs=j[-1]
						j=j[:-1]
					if j[0][0]==MenuMaker.checkBoxMarker:
						label=j[0][1:]
						kwargs['itemStateChanged']=j[1]
						item=apply(Config.factory.constructCheckboxMenuItem,
							(label,),kwargs)
						if len(j)>3:
							item.setState(j[3])
						if len(j)>4:
							item.setShortcut(
							 Config.factory.constructMenuShortcut(
							 	ord(j[4]),
								not(ord(j[4])&
								32)))
					else:
						label=j[0]
						kwargs['actionPerformed']=j[1]
						item=apply(Config.factory.constructMenuItem,
							(label,),kwargs)
						if len(j)>3:
							item.setShortcut(
							 Config.factory.constructMenuShortcut(
							 	ord(j[3]),
								not(ord(j[3])&
								32)))
					if len(j)>2:
						item.setEnabled(j[2])
					menuitems[label]=item
				menu.add(item)
		if len(i)>2:
			menu.setEnabled(i[2])
		menuitems[i[0]]=menu
		return menu


	def _makeDefaultToolBars(self):
		self.toolbars,self.buttons=self.makeToolBars(self.tBar)


	def _getAllComponents(self):
		ac=[]
		self._getAllCompsRec(self.frame,ac)
		return ac


	def _getAllCompsRec(self,cn,ac):
		ac.append(cn)
		try:
			for i in cn.getComponents():
				self._getAllCompsRec(i,ac)
		except:
			pass

###############################################################################

