#  $Id: TextLayer.py,v 2.13.1.3 2001-08-14 17:05:10-05 brinkman Exp $
#  An implementation of the layer interface (see LayeredFrame) for text.
#  Copyright (C) 2000  Peter Brinkmann (brinkman@math.uiuc.edu)
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
TextLayer.py: An implementation of the Layer interface (see LayeredFrame.py).

$Id: TextLayer.py,v 2.13.1.3 2001-08-14 17:05:10-05 brinkman Exp $
"""

import Config
from MenuMaker import MenuMaker
import string

###############################################################################

class TextLayer(MenuMaker):
	"""
$Id: TextLayer.py,v 2.13.1.3 2001-08-14 17:05:10-05 brinkman Exp $

An implementation of the layer interface (see LayeredFrame.py) that provides
basic functionality for handling text (in particular, there's a history, undo
functions, some flags indicating the current status of the data, etc.). See
the documentation strings of the various methods.

Needs to be subclassed to be really useful (this class doesn't define any
menu items or tool bars because this would contradict its abstract nature).
"""

	version="""
$Id: TextLayer.py,v 2.13.1.3 2001-08-14 17:05:10-05 brinkman Exp $
(c) 2000 Peter Brinkmann (brinkman@math.uiuc.edu)
"""
	helptext="""
Subclasses should override this variable with their own help text.
"""

###############################################################################

	def __init__(self,ms,dat='',prefwidth=80,prefheight=30,
			editable=1,name='TextLayer'):
		"""
Constructor for TextLayer.

Constructors of subclasses should call this method first.

Parameters:

	- ms: the master object of this instance
			(e.g., the ambient frame)

	- prefwidth: the preferred width of the TextArea

	- prefheight: the preferred height of the TextArea

	- editable: flag indicating whether the text is editable

Also, instances of this class have a flag called 'unsaved' that is nonzero
iff the current data has not been saved.
"""
		self.change=0
		self.unsaved=0
		self.data=None
		self.prefwidth=prefwidth
		self.prefheight=prefheight
		self.editable=editable
		self.history=[]
		self.hindex=-1
		self.searchIndex=-1
		self.searchString=None
		MenuMaker.__init__(self,ms,0,name)
		self.setData(dat)
		self.change=0
		self.unsaved=0
		self.updateWidgets()

###############################################################################

	def evalData(self,dat):
		"""
Needs to be overridden by subclasses. evalData evaluates new data and is
supposed to throw an exception if the new data is bad.
"""
		pass

###############################################################################

	def notEmpty(self):
		"""
Returns a nonzero value iff data is available (will frequently be overridden
by subclasses).
"""
		return (self.getText()!='' or self.data!='')

###############################################################################

	def find(self,s):
		"""
Searches a string in the current text, starting at the beginning. Highlights
the string if found.

Parameters:
	- s: search string

Returns nonzero value if and only if string has been found.
"""
		self.searchIndex=-1
		self.searchString=s
		return self.findAgain()

###############################################################################

	def findAgain(self,e=None):
		"""
Looks for the next occurrence of the current search string. Highlights the
string if found.

Parameters:
	- e: dummy parameter so this method can function as a call-back function

Returns nonzero value if and only if string has been found.
"""
		if self.searchString==None:
			raise Exception(_('No search string available.'))
		self.searchIndex=\
			string.find(self.getText(),self.searchString,self.searchIndex+1)
		if self.searchIndex<0:
			self.text.select(0,0)
			return 0
		self.text.select(self.searchIndex,
						self.searchIndex+len(self.searchString))
		self.updateWidgets()
		return 1

###############################################################################

	def replace(self,s1,s2):
		"""
Replaces all occurrences of s1 by s2.

Parameters:
	- s1: search string
	- s2: replace string
"""
		self.setDataSafely(string.replace(self.getText(),s1,s2))


###############################################################################
# In most cases, the remaining methods do not need to be overridden by        #
# subclasses!                                                                 #
###############################################################################

	def canUndo(self):
		"""
Returns a nonzero value iff there is a previous entry in the history (i.e.,
undo is possible).
"""
		return (self.hindex>0 or (self.hindex>=0
			 and (self.hasChanged() or
				 self.data!=self.history[self.hindex])))

###############################################################################

	def canRedo(self):
		"""
Returns a nonzero value iff there is a next entry in the history (i.e.,
undo has been performed recently.
"""
		return (self.hindex<(len(self.history)-1))

###############################################################################

	def hasChanged(self):
		"""
Returns a nonzero value iff the current text has not been checked in.
"""
		return self.change

###############################################################################

	def clear(self,e=None):
		"""
Erases the current text (can be undone).
"""
		if self.doubleCheck():
			self.setData('')
			self.change=0
			self.unsaved=0
			self.updateWidgets()

###############################################################################

	def doubleCheck(self):
		"""
Safety check to be performed before using the current data (asks the user
whether the current state should be checked in).

Returns nonzero value iff the method that called doubleCheck should proceed.
"""
		if not self.hasChanged():
			return 1
		NO=_('No')
		CANCEL=_('Cancel')
		ans=self.question(self.name,_('Check in current state?'),
					((_('Yes'),self.checkIn,1),
					 (NO,lambda:None),
					 (CANCEL,lambda:None)
					),
					1
				)
		if ans==NO:
			return 1
		if ans==CANCEL or ans==None:
			return 0
		return not self.hasChanged()

###############################################################################

	def setDataSafely(self,dat):
		"""
Sets self.data to dat after safety check. Throws an exception if data is bad.

Parameters:

	- dat: new contents of TextArea
"""
		if self.doubleCheck():
			self.setData(dat)

###############################################################################

	def setData(self,dat):
		"""
Sets self.data to dat. Throws an exception if data is bad. The new data becomes
part of the history (for undo, for example).

Parameters:

	- dat: new contents of TextArea
"""
		self.setDataWithoutHistory(dat)
		self.adjustHistory()
		self.change=0
		self.updateWidgets()

###############################################################################

	def undo(self,e=None):
		"""
Reverts to the most recent history entry.
"""
		if (self.hindex>=0):
			if self.hasChanged() or \
					self.data!=self.history[self.hindex]:
				self.setDataWithoutHistory(
					self.history[self.hindex])
				self.change=0
				self.updateWidgets()
			elif self.hindex>0:
				self.hindex=self.hindex-1
				self.setDataWithoutHistory(
					self.history[self.hindex])
				self.change=0
				self.updateWidgets()
			else:
				self.errorBox(_('End of history.'))

###############################################################################

	def redo(self,e=None):
		"""
Reverts to the next history entry.
"""
		if self.hindex<(len(self.history)-1):
			self.hindex=self.hindex+1
			self.setDataWithoutHistory(
				self.history[self.hindex])
			self.change=0
			self.updateWidgets()

###############################################################################

	def checkIn(self,e=None):
		"""
Attempts to check in new data and displays an error message if the new data
is bad.
"""
		self.setAllCursorsBusy()
		try:
			try:
				self.setData(self.getText())
			finally:
				self.resetAllCursors()
		except:
			self.exceptionBox()

###############################################################################

	def save(self,e=None):
		"""
Saves the current data as text.
"""
		if (not MenuMaker.isApplet) and self.doubleCheck():
			if self.notEmpty():
				self.saveDialog(self.name+': '+
					_('Save'),
					self.__saveIt__)
				self.unsaved=0
				self.updateWidgets()

###############################################################################

	def load(self,e=None):
		"""
Reads new data from a file.
"""
		if (not MenuMaker.isApplet) and self.doubleCheck():
			try:
				file=self.loadDialog(self.name+': '+
					_('Open'),
					self.__loadIt__)
				self.unsaved=0
				self.change=0
				self.updateWidgets()
			except:
				self.exceptionBox()

###############################################################################
# The remaining functions are for internal use only!                          #
###############################################################################

	def makeWidgets(self):
		self.text=Config.factory.constructTextArea(
			self.prefheight,self.prefwidth)
		self.text.setEditable(self.editable)
		self.text.setFont(Config.factory.textfont)
		self.text.keyTyped=self.handleChange
		self.panel=Config.factory.constructPanel(
			Config.factory.constructBorderLayout())
		self.panel.add(self.text,Config.factory.locations['C'])
		for i in self.toolbars.keys():
			self.panel.add(self.toolbars[i],
				Config.factory.locations[i])


	def handleChange(self,e=None):
		self.change=1
		self.unsaved=1
		self.updateWidgets()


	def adjustHistory(self):
		if self.notEmpty():
			if self.hindex<(len(self.history)-1):
				self.history=self.history[:(self.hindex+1)]
			self.history.append(self.data)
			self.hindex=self.hindex+1
		self.updateWidgets()


	def __saveIt__(self,fn):
		file=open(fn,'w')
		file.write(self.data)
		file.flush()
		file.close()


	def __loadIt__(self,fn):
		f=open(fn)
		dat=f.read()
		f.close()
		self.setData(dat)


	def finish(self):
		if self.unsaved and not MenuMaker.isApplet:
			CANCEL=_("Don't close")
			ans=self.question(self.name,
					_('Contents have not been saved!'),
						((_('Save now'),
							self.save,1),
						 (_("Don't save"),
							lambda:None),
						 (CANCEL,lambda:None)
						),
						1
					)
			if ans==CANCEL or ans==None:
				return 0
		try:
			self.panel=None
			self.text=None
		except:
			pass
		return 1


	def setText(self,s):
		self.text.setText(s)


	def getText(self):
		return self.text.getText()


	def setDataWithoutHistory(self,dat):
		self.evalData(dat)
		self.data=dat
		self.setText(self.data)
		self.change=1
		self.unsaved=1
		self.updateWidgets()

###############################################################################


