#!/usr/bin/env python
#  $Id: generator.py,v 2.27.1.6 2001-08-15 21:42:24-05 brinkman Exp $
#  A filter that automatically creates Jython GUIs from description files.
#  Copyright (C) 2000  Peter Brinkmann (brinkman@math.uiuc.edu)
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
generator.py: An automatic generator for graphical user interfaces.

$Id: generator.py,v 2.27.1.6 2001-08-15 21:42:24-05 brinkman Exp $
"""

import string
import sys
import StringIO
import getopt


class Struct:
	"""
Just a dummy class for emulating structs.
"""
	pass


class generator:

	"""
$Id: generator.py,v 2.27.1.6 2001-08-15 21:42:24-05 brinkman Exp $

This class provides a filter that reads GUI descriptions files whose
format has to conform to the following example:

------------------------------------------------------------------------

#begin gui
# example.dat: A sample GUI definition file.

# $Id: generator.py,v 2.27.1.6 2001-08-15 21:42:24-05 brinkman Exp $

# comments are allowed!

Name
	example			# any valid Python class name is ok
				# defaults to newgui

Init
	Default		# specifies configuration package, defaults to 'Default'
	
Type
	PowerFrame		# choices are MenuMaker, PowerFrame,
					# Layer, and TextLayer
					# defaults to PowerFrame

Menubar
	File tearoff		# possible markers: tearoff, enabled:<flag>,
						# extra:<dictionary>
				# defaults to enabled:1
		New		# cascading menu
			Window
			Copy	enabled:0
			Cascade2
				Spam	# yet another cascading menu
		---		# separator
		Open		# possible markers: enabled:<flag>
		Save
		---
		Close	self.close shortcut:c
					# can provide callback functions
					# beginning with 'self.'
					# keyboard accelerator <Ctrl>-c
		Exit	self.exit shortcut:X
					# keyboard accelerator <Ctrl>+<Shift>+x
	Edit	extra:{'font':self.textfont}
					# the extra tag creates a dictionary that will be passed
					# to the constructor as a keyword argument
					# this is also the place to put icons and such for java 1.2
		Undo	extra:{'font':self.textfontlarge}
		Redo		enabled:0
		Checkbox1	unchecked	# a checkbox: options are
						# checked or unchecked
		Checkbox2	checked enabled:0
		---
		Item_with_spaces # underscores ('_') will become spaces (' ')
	Test	enabled:0
		Item
		Another_item
		# the extra tag creates a dictionary that'll be passed to the
		# constructor as a keyword argument

Toolbar
	N			# tool bar at the top (north)
		Button1		# possible markers: enabled:<flag>
				# defaults to enabled:1
		Button2	enabled:0

	E
		Button3
		Button4
	W
		Button5 enabled:0
		Exit self.exit	 # callback function beginning with 'self.'
	S
		Button6
		Button7	enabled:0
		Button8		extra:{'font':self.textfont}
#end

------------------------------------------------------------------------

The output is a (working) jython program that implements the GUI described
in the above fashion.

"""

	COMMENT='#'
	ENABLED='enabled:'
	SHORTCUT='shortcut:'
	EXTRA='extra:'
	TEAROFF='tearoff'
	CHECKED='checked'
	UNCHECKED='unchecked'
	SEPARATOR='---'
	SELF='self.'
	PROTSPACE='_'
	SPACE=' '
	PF='PowerFrame'
	PFMOD='PowerFrame2'
	LY='Layer'
	TL='TextLayer'
	MM='MenuMaker'
	STARTBLOCK='#begin'
	ENDBLOCK='#end'
	GUITAG='gui'
	PREAMBLETAG='preamble'
	CLASSTAG='class'
	INITTAG='init'
	CONSTRUCTORTAG='constructor'
	WIDGETSTAG='widgets'
	UPDATETAG='update'
	CALLBACKTAG='callback'
	MISCTAG='misc'
	MAINTAG='main'
	

	def readInput(self,inp,code):
		code.gui='# put gui description here\n'
		code.preamble='# put preamble (imports and such) here\n'
		code.classdef=None
		code.oldname=None
		code.oldtype=None
		code.init='# put class level variables here\n'
		code.constructor=None
		code.widgets=None
		code.update='# put your updates for widgets here\n'
		code.callbacks={}
		code.misc='# define additional methods here\n'
		code.main=None
		while 1:
			line=inp.readline()
			if not line:
				return
			if line[:len(self.STARTBLOCK)]==self.STARTBLOCK:
				blk=self.readBlock(inp)
				lst=string.split(line)
				if lst[1]==self.GUITAG:
					code.gui=blk
				elif lst[1]==self.PREAMBLETAG:
					code.preamble=blk
				elif lst[1]==self.CLASSTAG:
					code.classdef=blk
					code.oldname=lst[2]
					code.oldtype=lst[3]
				elif lst[1]==self.INITTAG:
					code.init=blk
				elif lst[1]==self.CONSTRUCTORTAG:
					code.constructor=blk
				elif lst[1]==self.WIDGETSTAG:
					code.widgets=blk
				elif lst[1]==self.UPDATETAG:
					code.update=blk
				elif lst[1]==self.CALLBACKTAG:
					code.callbacks[lst[2]]=blk
				elif lst[1]==self.MISCTAG:
					code.misc=blk
				elif lst[1]==self.MAINTAG:
					code.main=blk


	def readBlock(self,inp):
		blk=''
		while 1:
			line=inp.readline()
			if not line or \
			   line[:len(self.ENDBLOCK)]==self.ENDBLOCK:
				return blk
			blk=blk+line


	def parseLines(self,s):
		lines=[]
		for line in string.split(s,'\n'):
			lst=string.split(string.split(line,self.COMMENT)[0])
			if len(lst)>0:
				lines.append((string.find(line,lst[0]),lst))
		return lines


	def parseTree(self,lines):
		tree=[]
		i=0
		while i<len(lines):
			j=i+1
			while j<len(lines) and lines[j][0]>lines[i][0]:
				j=j+1
			if j>i+1:
				tree.append([lines[i][1],
					self.parseTree(lines[i+1:j])])
			else:
				tree.append(lines[i][1])
			i=j
		return tree


	def evaluateTree(self,tree,code):
		code.name='newgui'
		code.config='Default'
		code.type=self.PF
		code.superclass=self.PF
		code.supermod=self.PFMOD
		mb=[]
		tb=[]
		for item in tree:
			if item[0][0]=='Menubar':
				mb=item[1]
			elif item[0][0]=='Toolbar':
				tb=item[1]
			elif item[0][0]=='Name':
				code.name=item[1][0][0]
			elif item[0][0]=='Init':
				code.config=item[1][0][0]
			elif item[0][0]=='Type':
				code.type=item[1][0][0]
				if code.type==self.LY:
					code.superclass=self.MM
					code.supermod=self.MM
				elif code.type==self.PF:
					code.superclass=self.PF
					code.supermod=self.PFMOD
				else:
					code.superclass=code.type
					code.supermod=code.type
				if code.type!=self.PF and \
				   code.type!=self.MM and \
				   code.type!=self.TL and \
				   code.type!=self.LY:
					raise Exception(
						'Bad type: '+`code.type`)
			else:
				raise Exception('Bad input:\n'+`item`)
		self.buildMenubar(mb,code)
		self.buildToolbar(tb,code)


	def buildMenubar(self,mb,code):
		code.mlist=[]
		code.mcallbacks=[]
		code.mbar=['(']
		for i in mb:
			self.buildMenubarRec(i,0,code)
		code.mbar.append(')')


	def buildMenubarRec(self,item,dpth,code):
		if type(item[0])==type(''):
			if dpth<1:
				raise Exception('Bad input:\n'+`item`)
			var,entry,callback,enabled,checked,shortcut,extra=\
					self.parseMenuItem(item,code)
			if var==None:
				code.mbar.append(entry)
			else:
				code.mbar.append('(')
				code.mbar.append(entry)
				code.mbar.append(',')
				code.mbar.append(callback)
				if shortcut!=None or checked!=None\
				   or enabled!='1':
					code.mbar.append(',')
					code.mbar.append(`(enabled=='1')`)
				if checked!=None:
					code.mbar.append(',')
					code.mbar.append(`checked`)
				if shortcut!=None:
					code.mbar.append(',')
					code.mbar.append("'"+shortcut+"'")
				if extra!=None:
					code.mbar.append(',')
					code.mbar.append(extra)
				code.mbar.append(')')
		else:
			var,enabled,tearoff,extra=self.parseMenu(item[0],code)
			code.mbar.append('(')
			code.mbar.append(var)
			code.mbar.append(',')
			code.mbar.append('(')
			for i in item[1]:
				self.buildMenubarRec(i,dpth+1,code)
			code.mbar.append(')')
			if tearoff!=0 or enabled!='1':
				code.mbar.append(',')
				if enabled!='1':
					code.mbar.append('0')
				else:
					code.mbar.append('1')
			if tearoff!=0:
				code.mbar.append(',')
				code.mbar.append(`tearoff`)
			if extra!=None:
				code.mbar.append(',')
				code.mbar.append(extra)
			code.mbar.append(')')
		code.mbar.append(',')


	def parseMenuItem(self,item,code):
		if item[0]==self.SEPARATOR:
			return None,'self.separator',None,None,None,None,None
		var='self.MI_%s'%item[0]
		entry=var
		callback='self.menuCallback_%s'%item[0]
		cbflag=1
		enabled='1'
		checked=None
		shortcut=None
		extra=None
		label=string.replace(item[0],self.PROTSPACE,self.SPACE)
		for i in item[1:]:
			if i[:len(self.ENABLED)]==self.ENABLED:
				enabled=i[len(self.ENABLED):]
			elif i[:len(self.SHORTCUT)]==self.SHORTCUT:
				shortcut=i[len(self.SHORTCUT):][0]
			elif i[:len(self.EXTRA)]==self.EXTRA:
				extra=i[len(self.EXTRA):]
			elif i==self.CHECKED:
				if checked==None:
					entry='self.checkBoxMarker+'+var
				checked=1
			elif i==self.UNCHECKED:
				if checked==None:
					entry='self.checkBoxMarker+'+var
				checked=0
			elif i[:len(self.SELF)]==self.SELF:
				callback=i
				cbflag=0
			else:
				raise Exception('Bad input:\n'+`item`)
		code.mlist.append([var,label,enabled])
		if cbflag!=0:
			code.mcallbacks.append([callback,label])
		return var,entry,callback,enabled,checked,shortcut,extra


	def parseMenu(self,item,code):
		var='self.MI_%s'%item[0]
		enabled='1'
		extra=None
		tearoff=0
		label=string.replace(item[0],self.PROTSPACE,self.SPACE)
		for i in item[1:]:
			if i[:len(self.ENABLED)]==self.ENABLED:
				enabled=i[len(self.ENABLED):]
			elif i[:len(self.EXTRA)]==self.EXTRA:
				extra=i[len(self.EXTRA):]
			elif i==self.TEAROFF:
				tearoff=1
			else:
				raise Exception('Bad input:\n'+`item`)
		code.mlist.append([var,label,enabled])
		return var,enabled,tearoff,extra


	def buildToolbar(self,tb,code):
		code.blist=[]
		code.bcallbacks=[]
		code.bbar=['(']
		for loc in tb:
			code.bbar.append('(')
			code.bbar.append("'%s'"%loc[0][0])
			code.bbar.append(',')
			code.bbar.append('(')
			for bt in loc[1]:
				var,callback,enabled,extra=self.parseButton(bt,code)
				code.bbar.append('(')
				code.bbar.append(var)
				code.bbar.append(',')
				code.bbar.append(callback)
				if enabled!='1':
					code.bbar.append(',')
					code.bbar.append('0')
				if extra!=None:
					code.bbar.append(',')
					code.bbar.append(extra)
				code.bbar.append(')')
				code.bbar.append(',')
			code.bbar.append(')')
			code.bbar.append(')')
			code.bbar.append(',')
		code.bbar.append(')')


	def parseButton(self,bt,code):
		var='self.BT_%s'%bt[0]
		callback='self.buttonCallback_%s'%bt[0]
		cbflag=1
		enabled='1'
		extra=None
		label=string.replace(bt[0],self.PROTSPACE,self.SPACE)
		for i in bt[1:]:
			if i[:len(self.ENABLED)]==self.ENABLED:
				enabled=i[len(self.ENABLED):]
			elif i[:len(self.EXTRA)]==self.EXTRA:
				extra=i[len(self.EXTRA):]
			elif i[:len(self.SELF)]==self.SELF:
				callback=i
				cbflag=0
		code.blist.append([var,label,enabled])
		if cbflag!=0:
			code.bcallbacks.append([callback,label])
		return var,callback,enabled,extra


	def generateCode(self,*arg):
		if len(arg)>0:
			inp=arg[0]
		else:
			inp=sys.stdin

		code=Struct()
		self.readInput(inp,code)
		self.evaluateTree(
			self.parseTree(self.parseLines(code.gui)),code)

		code.typeChange=(code.oldtype!=None and
				 code.oldtype!=code.type)
		code.nameChange=(code.oldname!=None and
				 code.oldname!=code.name)

		if len(arg)>1:
			out=arg[1]
		else:
			out=sys.stdout

		self.makePreamble(out,code)
		self.makeClassDef(out,code)
		self.makeStart(out,code)
		self.makeMakeWidgets(out,code)
		self.makeUpdateWidgets(out,code)
		self.makeCallbacks(out,code)
		self.makeMisc(out,code)
		self.makeMain(out,code)


	def makePreamble(self,out,code):
		out.write('__guistructure__="""\n%s %s\n'
					%(self.STARTBLOCK,self.GUITAG))
		out.write(code.gui)
		out.write('%s\n"""\n\n' % self.ENDBLOCK)
		out.write('\nimport Config\n')
		out.write('\nimport %s\n' % code.config)
		out.write('from %s import %s\n\n'%
			(code.supermod,code.superclass))
		out.write('%s %s\n' % (self.STARTBLOCK,self.PREAMBLETAG))
		out.write(code.preamble)
		out.write('%s\n\n'%self.ENDBLOCK)


	def makeClassDef(self,out,code):
		out.write('%s %s %s %s\n'%
			(self.STARTBLOCK,self.CLASSTAG,code.name,code.type))
		if code.classdef==None or code.typeChange or code.nameChange:
			out.write(
				('class %s(%s):\n' % (code.name,
							code.superclass)))
		else:
			out.write('%s'%code.classdef)
		out.write('%s\n'%self.ENDBLOCK)
		if code.classdef!=None and (code.typeChange or
						 code.nameChange):
			out.write('#\n# previous class definition\n')
			lst=string.split(code.classdef,'\n')
			for i in lst:
				out.write('# %s\n'%i)
		out.write('\n\n%s %s\n%s%s\n\n\n'%
			(self.STARTBLOCK,self.INITTAG,code.init,self.ENDBLOCK))

		out.write('%s %s\n'%
			(self.STARTBLOCK,self.CONSTRUCTORTAG))
		if code.constructor==None or code.typeChange:
			if code.type==self.PF:
				out.write(
					("\tdef __init__(self,ms=None,name='%s'):\n"
						% code.name)+
					'\t\t# a simple constructor. '+
					'Modify according to your needs!\n'+
					('\t\t%s.__init__(self,ms,name)\n' % code.type)+
					'\t\tself.updateWidgets()\n'+
					'\t\tself.setVisible(1)\n'+
					'\t\tself.pack()\n')
			elif code.type==self.LY:
				out.write(
				   ("\tdef __init__(self,ms=None,data=None,name='%s'):\n"
						% code.name)+
					'\t\t# a simple constructor. '+
					'Modify according to your needs!\n'+
					'\t\tMenuMaker.__init__(self,ms,0,name)\n'+
					'\t\tself.updateWidgets()\n')
			elif code.type==self.TL:
				out.write(
				   ("\tdef __init__(self,ms=None,data=None,name='%s'):\n"
						% code.name)+
					'\t\t# a simple constructor. '+
					'Modify according to your needs!\n'+
					'\t\tTextLayer.__init__('+
					'self,ms,data,80,30,1,name)\n'+
					'\t\tself.updateWidgets()\n')
			elif code.type==self.MM:
				out.write(
					'\t# constructor inherited '+
					'from MenuMaker\n')
		else:
			out.write(code.constructor)
		out.write('%s\n'%self.ENDBLOCK)
		if code.constructor!=None and code.typeChange:
			out.write('#\n# previous constructor\n')
			lst=string.split(code.constructor,'\n')
			for i in lst:
				out.write('# %s\n'%i)
		out.write('\n\n')


	def makeStart(self,out,code):
		out.write(
			'\tdef start(self):\n\t\t# Modify at your own risk!\n')
		for i in code.mlist:
			out.write("\t\t%s=_('%s')\n"%(i[0],i[1]))
		cnt=1
		step='  '
		out.write('\t\tself.mBar=\\')
		for i in code.mbar:
			if i=='(':
				out.write('\n\t\t'+step*cnt)
				cnt=cnt+1
			elif i==')':
				cnt=cnt-1
				out.write('\n\t\t'+step*cnt)
			out.write(i)
		out.write('\n')
		for i in code.blist:
			out.write("\t\t%s=_('%s')\n"%(i[0],i[1]))
		out.write('\t\tself.tBar=\\')
		cnt=1
		for i in code.bbar:
			if i=='(':
				out.write('\n\t\t'+step*cnt)
				cnt=cnt+1
			elif i==')':
				cnt=cnt-1
				out.write('\n\t\t'+step*cnt)
			out.write(i)
		out.write('\n\n\n')


	def makeMakeWidgets(self,out,code):
		out.write('%s %s\n'%
			(self.STARTBLOCK,self.WIDGETSTAG))
		if code.widgets==None or code.typeChange:
			if code.type==self.PF:
				out.write('\tdef makeWidgets(self):\n')
				out.write(
				   '\t\t# modify this to include contents\n'+
				   '\t\tself.add(Config.factory.constructLabel('+
					"'Put contents here!'),\n"+
				   "\t\t\tConfig.factory.locations['C'])\n")
			elif code.type==self.LY:
				out.write('\tdef makeWidgets(self):\n')
				out.write(
				 "\t\t# don't modify the next few lines\n"+
				 '\t\tself.panel=Config.factory.constructPanel('+
				 'Config.factory.constructBorderLayout())\n'+
					'\t\tfor i in self.toolbars.keys():\n'+
				   '\t\t\tself.panel.add(self.toolbars[i],'+
					'Config.factory.locations[i])\n\n'+
				 '\t\t# modify this part to include contents\n'+
					'\t\tself.panel.add('+
				   "Config.factory.constructLabel("+
				   "'Put contents here!'),\n"+
				   "\t\t\tConfig.factory.locations['C'])\n")
			elif code.type==self.TL:
				out.write(
				   '# put code for creating widgets here\n')
			elif code.type==self.MM:
				out.write(
				   '# put code for creating widgets here\n')
		else:
			out.write(code.widgets)
		out.write('%s\n'%self.ENDBLOCK)
		if code.widgets!=None and code.typeChange:
			out.write('#\n# previous makeWidgets\n')
			lst=string.split(code.widgets,'\n')
			for i in lst:
				out.write('# %s\n'%i)
		out.write('\n\n')


	def makeUpdateWidgets(self,out,code):
		out.write(
			'\tdef updateWidgets(self,e=None):\n')
		if len(code.mlist)>0 or len(code.blist)>0:
			for i in code.mlist:
				out.write(
				 "\t\tself.menuitems["+
				 "%s].setEnabled(%s) # enable '%s'\n" 
				   % (i[0],i[2],i[1]))
			for i in code.blist:
				out.write(
				 "\t\tself.buttons["+
				 "%s].setEnabled(%s) # enable '%s'\n"
				   % (i[0],i[2],i[1]))
		else:
			out.write('\t\tpass\n')
		out.write('%s %s\n'%(self.STARTBLOCK,self.UPDATETAG))
		out.write(code.update)
		out.write('%s\n\n\n'%self.ENDBLOCK)


	def makeCallbacks(self,out,code):
		for i in code.mcallbacks:
			cbname=i[0][5:]
			out.write('%s %s %s\n'%
				(self.STARTBLOCK,self.CALLBACKTAG,
					cbname))
			if not (cbname in code.callbacks.keys()):
				out.write('\tdef %s(self,e=None):'
							%cbname)
				out.write(
					("\n\t\t# callback for '%s'\n" % i[1])+
					("\t\tself.infoBox('%s'," % i[1])+
					("'You clicked on %s!'" % i[1])+
					",'OK',0)\n")
			else:
				out.write(code.callbacks[cbname])
				del code.callbacks[cbname]
			out.write('%s\n\n\n'%self.ENDBLOCK)

		for i in code.bcallbacks:
			cbname=i[0][5:]
			out.write('%s %s %s\n'%
				(self.STARTBLOCK,self.CALLBACKTAG,
					cbname))
			if not (cbname in code.callbacks.keys()):
				out.write('\tdef %s(self,e=None):'
					%cbname)
				out.write(
					("\n\t\t# callback for '%s'\n" % i[1])+
					("\t\tself.infoBox('%s'," % i[1])+
					("'You clicked on %s!'" % i[1])+
					",'OK',0)\n")
			else:
				out.write(code.callbacks[cbname])
				del code.callbacks[cbname]
			out.write('%s\n\n\n'%self.ENDBLOCK)

		if len(code.callbacks)>0:
			out.write('# obsolete callback functions\n#\n')
			for i in code.callbacks.keys():
				lns=string.split(code.callbacks[i],'\n')
				for i in lns:
					out.write('# %s\n'%i)
			out.write('\n\n')


	def makeMisc(self,out,code):
		out.write('%s %s\n'%(self.STARTBLOCK,self.MISCTAG))
		if code.misc!=None:
			out.write(code.misc)
		else:
			out.write('# put miscellaneous functions '+
						'here\n')
		out.write('%s\n\n\n'%self.ENDBLOCK)


	def makeMain(self,out,code):
		out.write('%s %s\n'%(self.STARTBLOCK,self.MAINTAG))
		if code.main==None or code.typeChange or code.nameChange:
			if code.type==self.PF:
				out.write(
					'# a simple main routine\n'+
					"if __name__=='__main__':\n"+
					'\t%s()\n' % code.name)
			elif code.type==self.LY or code.type==self.TL:
				out.write(
					'# a simple main routine\n'+
					"if __name__=='__main__':\n"+
				   '\tfrom LayeredFrame2 import LayeredFrame\n'+
				   ('\tf=LayeredFrame(%s)\n' % code.name))
		else:
			out.write(code.main)
		out.write('%s\n'%self.ENDBLOCK)
		if code.main!=None and (code.typeChange or code.nameChange):
			out.write('#\n# previous main routine\n')
			lst=string.split(code.main,'\n')
			for i in lst:
				out.write('# %s\n'%i)
		out.write('\n\n')


def usage(str,name):
	str.write(
"""Usage: %s [-h] [--help] [-o file] [--output file] [file]

By default, input comes from stdin and output goes to stdout.
If a file name is present, input is read from this file, and
the contents of the file are replaced by the output, unless
a separate output file is specified with -o or --output.
""" % name)


def main(argv):
	infile=None
	outfile=None

	try:
		opts,args=getopt.getopt(argv[1:],'ho:',['help','output='])
	except:
		usage(sys.stderr,argv[0])
		raise Exception('Unknown command line option.')

	if len(args)>0:
		if len(args)>1:
			usage(sys.stderr,argv[0])
			raise Exception('Too many arguments.')
		infile=args[0]
		outfile=infile

	for o,a in opts:
		if o in ('-o','--output'):
			outfile=a
		elif o in ('-h','--help'):
			usage(sys.stdout,argv[0])
			return

	if infile!=None and infile!='-':
		inp=open(infile,'r')
	else:
		inp=sys.stdin

	if outfile!=None and outfile!='-' and outfile!='':
		out=StringIO.StringIO()
	else:
		out=sys.stdout

	generator().generateCode(inp,out)

	if outfile!=None and outfile!='-' and outfile!='':
		f=open(outfile,'w')
		f.write(out.getvalue())
		f.flush()


if __name__=='__main__':
	main(sys.argv)

